"""
Update release numbers in various places, according to a release.ini file places at the project root.

For now, the following sections are supported:

+ Main file (__init__.py, Django settings file, etc.)
+ Sphinx conf.py file
+ sonar-project.properties
+ ansible vars file
+ node package.json file
+ setup.cfg
+ setup.py
+ pyproject.toml

"""
import configparser
import logging
import sys
from configparser import ConfigParser
from pathlib import Path
from typing import Optional

import click

from bump_release import helpers

# region Globals
__version__ = VERSION = "0.10.6"
RELEASE_FILE: Optional[Path] = None
RELEASE_CONFIG: Optional[ConfigParser] = None


# endregion Globals


@click.command()
@click.option(
    "-r",
    "--release-file",
    "release_file",
    help="Release file path, default `./release.ini`",
    type=click.Path(exists=True),
    default="release.ini",
)
@click.option(
    "-n",
    "--dry-run",
    "dry_run",
    is_flag=True,
    help="If set, no operation are performed on files",
    default=False,
)
@click.option(
    "-d",
    "--debug",
    "debug",
    is_flag=True,
    help="If set, more traces are printed for users",
    default=False,
)
@click.version_option(version=__version__)
@click.argument("release")
def bump_release(release: str, release_file: Optional[str] = None, dry_run: bool = False, debug: bool = False) -> int:
    """
    Update release numbers in various places, according to a release.ini file places at the project root.

    \b
    For now, the following sections are supported:

    \b
    + Main file (__init__.py, Django settings file, etc.)
    + Sphinx conf.py file
    + sonar-project.properties
    + ansible vars file
    + node package.json file
    + setup.cfg
    + setup.py
    + pyproject.toml (Require using `pip install bump_release[toml]`)

    \f
    :param release: Release number
    :param release_file: Release file path, default `./release.ini`
    :param dry_run: If `True`, no operation performed
    :param debug: If `True`, more traces are printed for users
    :return: 0 if success, 1|2 if error
    """
    # Loads the release.ini file
    global RELEASE_CONFIG, RELEASE_FILE

    if release_file is None:
        RELEASE_FILE = Path.cwd() / "release.ini"
    else:
        RELEASE_FILE = Path(release_file)

    if not RELEASE_FILE.exists():
        print(f"Unable to find release.ini file in the current directory {Path.cwd()}", file=sys.stderr)
        return 1

    RELEASE_CONFIG = helpers.load_release_file(release_file=RELEASE_FILE)
    try:
        return process_update(release_file=RELEASE_FILE, release=release, dry_run=dry_run, debug=debug)
    except Exception as e:
        print(f"ERROR: {e}", file=sys.stderr)
        return 2


def process_update(release_file: Path, release: str, dry_run: bool, debug: bool = False) -> int:
    """
    Processes the update of the release number.

    :param release_file: Release file path
    :param release: Release number
    :param dry_run: If `True`, no operation performed
    :param debug: If `True`, more traces are printed for users
    :return: 0 if success, 1|2 if error
    """
    version = helpers.VersionNumber(version=release)

    # Initialize the logging
    if debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # region Updates the main project (DJANGO_SETTINGS_MODULE file for django projects, __init__.py file...)
    try:
        new_row = update_main_file(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `main_project`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `main_project`: {e}")
    # endregion

    # region Updates sonar-scanner properties
    try:
        new_row = update_sonar_properties(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `sonar`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `sonar`: {e}")
    # endregion

    # region Updates setup.py file
    try:
        new_row = update_setup_file(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `setup`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `setup`: {e}")
    # endregion

    # region Update setup.cfg file
    try:
        new_row = update_setup_cfg_file(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `setup_cfg`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `setup`: {e}")
    # endregion Update setup.cfg file

    # region Updates sphinx file
    try:
        new_row = update_docs_conf(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `docs`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `docs`: {e}")
    # endregion

    # region Updates node packages file
    try:
        logging.debug("process_update() `node_module`")
        new_row = update_node_package(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(
                f"process_update() `node`: new_row = {new_row}",
            )
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `node`: {e}")
    # endregion

    # region Updates YAML file
    try:
        new_row = update_ansible_vars(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `ansible`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `ansible`: {e}")
    # endregion

    # region Updates pyproject.toml file
    try:
        new_row = update_pyproject(version=version, base_path=release_file.parent, dry_run=dry_run)
        if new_row is not None:
            logging.debug(f"process_update() `pyproject.toml`: new_row = {new_row.strip()}")
    except helpers.NothingToDoException as e:
        logging.warning(f"process_update() No release section for `pyproject.toml`: {e}")
    except helpers.UpdateException:  # noqa: E722
        logging.exception("process_update() `pyproject.toml`: UpdateException - tomli / tomli_w might be missing")

    # endregion Updates pyproject.toml file

    # region Updates the release.ini file with the new release number
    new_row = update_release_ini(path=release_file, version=version, dry_run=dry_run)
    if new_row is not None:
        logging.warning(f"process_update() `release.ini`: new_row = {new_row.strip()}")
    # endregion

    return 0


def update_main_file(version: helpers.VersionNumber, base_path: Path, dry_run: bool = True) -> Optional[str]:
    """
    Updates the main django settings file, or a python script with a __init__.py file.

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    if not RELEASE_CONFIG.has_section("main_project"):
        raise helpers.NothingToDoException("No `main_project` section in release.ini file")

    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("main_project", "path"), release_file_path=base_path)
        logging.debug(f"update_main_file() path = {path}")
        pattern = RELEASE_CONFIG["main_project"].get("pattern", "").strip('"') or helpers.MAIN_PROJECT_PATTERN
        template = RELEASE_CONFIG["main_project"].get("template", "").strip('"') or helpers.MAIN_PROJECT_TEMPLATE
    except configparser.Error as e:
        raise helpers.NothingToDoException("Unable to update main project file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_setup_file(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the setup.py file.

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    if not RELEASE_CONFIG.has_section("setup"):
        raise helpers.NothingToDoException("No `setup` section in release.ini file")

    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("setup", "path"), release_file_path=base_path)
        logging.debug(f"update_setup_file() path = {path}")
        pattern = RELEASE_CONFIG["setup"].get("pattern", "").strip('"') or helpers.SETUP_PATTERN
        template = RELEASE_CONFIG["setup"].get("template", "").strip('"') or helpers.SETUP_TEMPLATE

    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for setup file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_setup_cfg_file(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Update the setup.cfg file.

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    if not RELEASE_CONFIG.has_section("setup_cfg"):
        raise helpers.NothingToDoException("No `setup_cfg` section in release.ini file")

    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("setup_cfg", "path"), release_file_path=base_path)
        logging.debug(f"update_setup_cfg_file() path = {path}")
        pattern = RELEASE_CONFIG["setup_cfg"].get("pattern", "").strip('"') or helpers.SETUP_CFG_PATTERN
        template = RELEASE_CONFIG["setup_cfg"].get("template", "").strip('"') or helpers.SETUP_CFG_TEMPLATE

    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for setup.cfg file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_sonar_properties(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the sonar-project.properties file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    if not RELEASE_CONFIG.has_section("sonar"):
        raise helpers.NothingToDoException("No `sonar` section in release.ini file")

    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("sonar", "path"), release_file_path=base_path)
        logging.debug(f"update_sonar_properties() path = {path}")
        pattern = RELEASE_CONFIG["sonar"].get("pattern", "").strip('"') or helpers.SONAR_PATTERN
        template = RELEASE_CONFIG["sonar"].get("template", "").strip('"') or helpers.SONAR_TEMPLATE
    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for sonar file", e)
    return helpers.update_file(path=path, pattern=pattern, template=template, version=version, dry_run=dry_run)


def update_docs_conf(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the Sphinx conf.py file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    if not RELEASE_CONFIG.has_section("docs"):
        raise helpers.NothingToDoException("No `docs` section in release.ini file")

    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("docs", "path"), release_file_path=base_path)
        logging.debug(f"update_docs_conf() path = {path}")

        pattern_release = RELEASE_CONFIG["docs"].get("pattern_release", "").strip('"') or helpers.DOCS_RELEASE_PATTERN
        template_release = RELEASE_CONFIG["docs"].get("template_release", "").strip('"') or helpers.DOCS_RELEASE_FORMAT
        pattern_version = RELEASE_CONFIG["docs"].get("pattern_version", "").strip('"') or helpers.DOCS_VERSION_PATTERN
        template_version = RELEASE_CONFIG["docs"].get("template_version", "").strip('"') or helpers.DOCS_VERSION_FORMAT

    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for docs file", e)

    update_release = helpers.update_file(
        path=path,
        pattern=pattern_release,
        template=template_release,
        version=version,
        dry_run=dry_run,
    )
    update_version = helpers.update_file(
        path=path,
        pattern=pattern_version,
        template=template_version,
        version=version,
        dry_run=dry_run,
    )
    return str(update_release) + str(update_version)


def update_node_package(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the nodejs package file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        if "node" in RELEASE_CONFIG:
            path = helpers.resolve_path(path=RELEASE_CONFIG.get("node", "path"), release_file_path=base_path)
        elif "node_module" in RELEASE_CONFIG:
            path = helpers.resolve_path(path=RELEASE_CONFIG.get("node_module", "path"), release_file_path=base_path)
        else:
            raise helpers.NothingToDoException("No `node` or `node_module` section in release.ini file")
        logging.debug(f"update_node_package() path = {path}")
        key = RELEASE_CONFIG.get("node_module", "key", fallback=helpers.NODE_KEY)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for node packages file", e)
    return helpers.update_json_file(path=path, version=version, key=key, dry_run=dry_run)


def update_ansible_vars(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the ansible project variables file with the new release number

    :param version: Release number tuple (major, minor, release)
    :param base_path: Base search path
    :param dry_run: If `True`, no operation performed
    :return: changed string
    """
    assert RELEASE_CONFIG is not None
    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("ansible", "path"), release_file_path=base_path)
        logging.debug(f"update_ansible_vars() path = {path}")
        key = RELEASE_CONFIG.get("ansible", "key", fallback=helpers.ANSIBLE_KEY)  # noqa
    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for `ansible` file", e)
    return helpers.updates_yaml_file(path=path, version=version, key=key, dry_run=dry_run)


def update_pyproject(version: helpers.VersionNumber, base_path: Path, dry_run: bool = False) -> Optional[str]:
    """
    Updates the pyproject.toml file with the new release number

    :param version: release number, as (<major>, <minor>, <release>)
    :param base_path: Base search path
    :param dry_run: If `True`, the operation WILL NOT be performed
    :return: Updated lines
    """
    assert RELEASE_CONFIG is not None
    try:
        path = helpers.resolve_path(path=RELEASE_CONFIG.get("pyproject", "path"), release_file_path=base_path)
        logging.debug(f"update_pyproject() path = {path}")
        key = RELEASE_CONFIG.get("pyproject", "key", fallback=helpers.PYPROJECT_KEY)  # noqa
    except KeyError as ke:
        raise helpers.NothingToDoException("No pyproject key found in ini file for pyproject.toml file", ke)
    except configparser.Error as e:
        raise helpers.NothingToDoException("No action to perform for pyproject.toml file", e)

    # Update pyproject.toml file
    return helpers.update_toml_file(
        path=path,
        version=version,
        key=key,
        dry_run=dry_run,
    )


def update_release_ini(path: Path, version: helpers.VersionNumber, dry_run: bool = False) -> Optional[str]:
    """
    Updates the release.ini file with the new release number

    :param path: Release file path
    :param version: release number, as (<major>, <minor>, <release>)
    :param dry_run: If `True`, the operation WILL NOT be performed
    :return: Updated lines
    """
    return helpers.update_file(
        path=path,
        pattern=helpers.RELEASE_INI_PATTERN,
        template=helpers.RELEASE_INI_TEMPLATE,
        version=version,
        dry_run=dry_run,
    )



"""
__main__ for :mod:`bump_release` application

:creationdate: 17/09/2019 11:52
:moduleauthor: François GUÉRIN <fguerin@ville-tourcoing.fr>
:modulename: bump_release.__main__

"""
import sys

import bump_release

__author__ = "fguerin"


def main(*args):
    return bump_release.bump_release()


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



"""
helpers for :mod:`bump_release` application

:creationdate: 16/09/2019 08:18
:moduleauthor: François GUÉRIN <fguerin@ville-tourcoing.fr>
:modulename: bump_release.helpers

"""
import configparser
import json
import logging
import os
import re
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

__author__ = "fguerin"

RELEASE_CONFIG = None
BASE_DIR = os.getcwd()
# region Constants
# Node (JSON value update)
NODE_PACKAGE_FILE: str = "package.json"
NODE_KEY: str = "version"

# main (re search and replace)
MAIN_PROJECT_PATTERN: str = r"^__version__\s*=\s*VERSION\s*=\s*['\"]([\d\w]+)\.([\d\w]+)\.([\d\w]+)['\"]$"
MAIN_PROJECT_TEMPLATE: str = '__version__ = VERSION = "{major}.{minor}.{release}"'

# Ansible yaml key
ANSIBLE_KEY: str = "git.version"

# sonar (re search and replace)
SONAR_PATTERN: str = r"^sonar.projectVersion=([\d\w]+\.[\d\w]+)$"
SONAR_TEMPLATE: str = "sonar.projectVersion={major}.{minor}"

# setup.py file
SETUP_PATTERN: str = r"^\s*version=['\"]([\d\w]+)\.([\d\w]+)\.([\d\w]+)['\"],$"
SETUP_TEMPLATE: str = '    version="{major}.{minor}.{release}",'

# setup.cfg file
SETUP_CFG_PATTERN: str = r"^version = ([\d\w]+)\.([\d\w]+)\.([\d\w]+)$"
SETUP_CFG_TEMPLATE: str = "version = {major}.{minor}.{release}"

# Sphinx (re search and replace)
DOCS_VERSION_PATTERN: str = r"^version\s*=\s*[\"'](.*)[\"']$"
DOCS_VERSION_FORMAT: str = 'version = "{major}.{minor}"'
DOCS_RELEASE_PATTERN: str = r"^release\s*=\s*[\"'](.*)[\"']$"
DOCS_RELEASE_FORMAT: str = 'release = "{major}.{minor}.{release}"'

RELEASE_INI_PATTERN: str = r"^current_release\s*=\s*['\"]?([\d\w]+)\.([\d\w]+)\.([\d\w]+)['\"]?$"
RELEASE_INI_TEMPLATE: str = "current_release = {major}.{minor}.{release}"

# pyproject.toml file
PYPROJECT_KEY = "project.version"

# endregion Constants
logger = logging.getLogger(__name__)


@dataclass
class VersionNumber:
    """Release number."""

    major: int
    minor: int
    release: int

    def __init__(self, **kwargs):
        """Initialize the release number"""
        super().__init__()
        if "version" in kwargs:
            self.major, self.minor, self.release = kwargs["version"].split(".")
        elif "major" in kwargs and "minor" in kwargs and "release" in kwargs:
            self.major = kwargs.get("major", 0)
            self.minor = kwargs.get("minor", 0)
            self.release = kwargs.get("release", 0)
        else:
            raise ValueError(f"Invalid version number: {kwargs}")

    def __str__(self):
        """Get the string representation of the release number"""
        return f"{self.major}.{self.minor}.{self.release}"

    def __repr__(self):
        """Get the representation of the release number"""
        return f"{self.major}.{self.minor}.{self.release}"


def load_release_file(release_file: Union[Path, str]) -> configparser.ConfigParser:
    """
    Loads the release.ini file, and parse the configuration.

    :param release_file: Path to the release file
    :return: Loaded config
    """
    release_config = configparser.ConfigParser()
    release_config.read(release_file)
    return release_config


def update_file(
    path: Path,
    pattern: str,
    template: str,
    version: VersionNumber,
    dry_run: Optional[bool] = False,
) -> Optional[str]:
    """
    Performs the **real** update of the `path` files, aka. replaces the row matched
    with `pattern` with `version_format` formatted according to `release`.

    :param path: path of the file to update
    :param pattern: regexp to replace
    :param template: release format
    :param version: Release number tuple (major, minor, release)
    :param dry_run: If `True`, no operation performed
    :return: New row
    """
    logger.debug(f"Updating {pattern}")
    version_re = re.compile(pattern)

    old_row, new_row = None, None
    counter = None
    with path.open(mode="r") as ifile:
        content_lines = ifile.readlines()
    new_content = deepcopy(content_lines)
    found = False
    for counter, row in enumerate(content_lines):
        searched = version_re.search(row)
        if searched:
            logging.debug(f"update_file({path}) a *MATCHING* row has been found:\n{counter} {row.strip()}")
            found = True
            old_row = deepcopy(row)
            new_row = template.format(major=version.major, minor=version.minor, release=version.release)
            if old_row.endswith("\r\n"):
                new_row += "\r\n"
            elif old_row.endswith("\r"):
                new_row += "\r"
            elif old_row.endswith("\n"):
                new_row += "\n"
            break

    if not found:
        raise ValueError(f"No row matching {pattern} found in {path} file.")

    if old_row and new_row:
        logging.info(f"update_file({path}) old_row:\n{old_row.strip()}\nnew_row:\n{new_row.strip()}")

    if dry_run:
        logging.info(
            f"update_file({path}) No operation performed, dry_run = {dry_run}",
        )
        return new_row

    if new_row and counter is not None:
        new_content[counter] = new_row
        with path.open(mode="w") as output_file:
            output_file.writelines(new_content)
        logging.info(f"update_file({path}) File updated.")
        return new_row

    raise UpdateException(f"An error has append on updating release for file {path}")


def update_json_file(
    path: Path,
    version: VersionNumber,
    key: str = NODE_KEY,
    dry_run: bool = False,
) -> str:
    """
    Updates the package.json file

    :param path: Node root directory
    :param version: Release number
    :param dry_run: If `True`, no operation performed
    :param key: json dict key (default: "release")
    :return: Nothing
    """
    try:
        with path.open(mode="r") as package_file:
            package = json.loads(package_file.read())
        new_package = deepcopy(package)
        new_package[key] = str(version)
        updated = json.dumps(new_package, indent=4)
        if not dry_run:
            with path.open(mode="w") as package_file:
                package_file.write(updated)
        return updated
    except IOError as ioe:
        raise UpdateException(f"update_json_file() Unable to perform `package_file` update: {ioe}")


class MyYAML(YAML):
    """Wrapper around ruamel.yaml to output strings directly."""

    def dump(self, data, stream=None, **kw):
        """Dump the data to the stream."""
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


def updates_yaml_file(
    path: Path,
    version: VersionNumber,
    key: str = ANSIBLE_KEY,
    dry_run: bool = False,
) -> str:
    """
    Replaces the version number in a YAML file, aka. ansible vars files

    :param path: Path to the yaml file
    :param version: New version to apply, as a tuple (major, minor, release)
    :param key: key in the files, as xxx.yyy
    :param dry_run: If True, no action is performed
    :returns: new file content
    """
    yaml = MyYAML()
    with path.open(mode="r") as vars_file:
        document = yaml.load(vars_file)
    internal_dict_value = document
    split_key = key.split(".")
    print(f"DEBUG: {split_key=}")
    if len(split_key) == 1:
        internal_dict_value.update({key: str(version)})
    else:
        for _key in split_key:
            internal_dict_value = internal_dict_value.get(_key)
            if internal_dict_value and _key == split_key[-1]:
                internal_dict_value = str(version)
    logging.info(f"updates_yml_file({vars_file}) node value = {internal_dict_value}")
    # Save the content
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.version = (1, 2)
    new_content = yaml.dump(document)
    if not dry_run:
        with path.open(mode="w") as vars_file:
            vars_file.write(new_content)
    return new_content


def update_toml_file(path: Path, version: VersionNumber, key: str = PYPROJECT_KEY, dry_run: bool = False) -> str:
    """
    Update a toml file with the new version.

    :param path: Path to the toml file
    :param version: New version to apply, as a tuple (major, minor, release)
    :param key: key in the files, as xxx.yyy
    :param dry_run: If True, no action is performed
    :returns: new file content
    """
    try:
        import tomli
        import tomli_w
    except ImportError as ie:
        raise UpdateException(f"update_toml_file() Unable to perform {path} update: tomli not found", ie)
    with path.open(mode="r") as toml_file:
        document = tomli.loads(toml_file.read())

    # Update the version
    split_key = key.split(".")

    node = document
    for _key in split_key:
        if _key == split_key[-1]:
            node.update({_key: str(version)})
        node = node.get(_key)

    logging.info(f"update_toml_file({path}) key: '{key}' / value: '{str(version)}'")
    # Write the updated file
    new_content = tomli_w.dumps(document)
    if not dry_run:
        with path.open(mode="w") as toml_file:
            toml_file.write(new_content)
    return new_content


class UpdateException(Exception):
    """
    An error has occurred during the release updating
    """

    pass


class NothingToDoException(UpdateException):
    """
    An error has occurred during the release updating: Nothing to do...
    """

    pass


def resolve_path(path: str, release_file_path: Path) -> Path:
    """
    Resolve a path to a `Path` object.
    """
    assert path, "A path must be provided"
    assert release_file_path, "A release file path must be provided"

    _path = Path(path)
    if _path.is_absolute():
        return _path.resolve()
    else:
        if release_file_path.is_dir():
            return release_file_path.joinpath(_path).resolve()
        else:
            return release_file_path.parent.joinpath(_path).resolve()
